/* GNU/Linux on ARM target machine support for single stepping without ptrace.
   Copyright (C) 1988, 1989, 1991, 1992, 1993, 1995-1999
   Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#ifdef NO_SINGLE_STEP

/* Code to support single stepping.  Arm Linux seems to have a problem
   with single stepping via the kernel ptrace() call with kernel version 
   2.0.31 and 2.0.35.

/* Non-zero if we just simulated a single-step ptrace call.  This is
   needed because we cannot remove the breakpoints in the inferior
   process until after the `wait' in `wait_for_inferior'. */
int one_stepped;

/* Arrays to hold the single step break locations and instructions
   read from memory and replaced by the breakpoint instruction. */
static CORE_ADDR step_break_locations[2];
static unsigned long step_shadow_insns[2];

/* Get the value from register `rn' in the opcode of an ARM
   data processing instruction. */
static unsigned long
opcode_getrn (unsigned long opcode)
{
  unsigned int reg = (opcode >> 16) & 0x0000000f;
  unsigned long val;

  if (PC_REGNUM == reg)
    val = ADDR_BITS_REMOVE (read_pc ());
  else
    val = read_register (reg);

  return val;
}

/* Get value of operand 2 in the opcode for an ARM ldr instruction. */
static unsigned long
opcode_getldrop2 (unsigned long opcode)
{
  unsigned long val;
  int shift, type;

  val = read_register (opcode & 0x0000000f);
  shift = (opcode >> 7) & 0x000001f;
  type = (opcode >> 5) & 0x00000003;

  switch (type)
    {
    case 0:
      val <<= shift;
      break;
    case 1:
      val >>= shift;
      break;

    case 2:
      val = (((signed long) val) >> shift);
      break;

    case 3:
    __asm__ __volatile__ ("mov %0, %0, ror %1": "=r" (val):"0" (val), "r" (shift));
      break;
    }

  return val;
}


/* Return true (non-zero) if Rd in opcode for a data processing instruction
   is the program counter. */
static int
opcode_is_dest_pc (unsigned long opcode)
{
  unsigned int reg = (opcode >> 12) & 0x0000000f;
  return (PC_REGNUM == reg ? 1 : 0);
}

/* Get the value of operand 2 in the opcode for an ARM ALU instruction. */
static unsigned long
opcode_getaluop2 (unsigned long opcode)
{
  unsigned val;
  int shift, type, Rn, PCOffset;

  Rn = (opcode >> 16) & 0x0000000f;
  PCOffset = 0;

  if (opcode & 0x00000001 << 25)
    {
      /* immediate operand */
      val = opcode & 0x000000ff;
      shift = (opcode >> 8) & 0x0000000f;
      type = 3;
    }
  else
    {
      /* register operand */
      int Rm = opcode & 0x0000000f;

      /* Is PC use valid for Rn and Rm? 
         If bit 4 of the opcode is 1, then the results of using the program 
         counter for Rn or Rm are unpredictable. */
      int PCValid = ((opcode >> 4) & 0x00000001) == 0;

      /* Warn if invalid program counter use detected. */
      if (((PC_REGNUM == Rm) || (PC_REGNUM == Rn)) && !PCValid)
	printf_filtered ("Invalid use of the program counter for Rn or Rm.\n");

      val = read_register (Rm);
      if (PCValid && (PC_REGNUM == Rm))
	val += 8;

      /* if Rn is the program counter we actually need to add 8 to the contents 
         of the program counter when calculating Rd. Make such an adjustment
         to the return value if necessary. */
      if (PCValid && (PC_REGNUM == Rn))
	PCOffset += 8;

      if (opcode & (1 << 4))
	shift = (int) read_register ((opcode >> 8) & 0x0000000f);
      else
	shift = (opcode >> 7) & 0x0000001f;

      type = (opcode >> 5) & 0x00000003;
    }

  switch (type)
    {
    case 0:
      val <<= shift;
      break;
    case 1:
      val >>= shift;
      break;
    case 2:
      val = (((signed long) val) >> shift);
      break;
    case 3:
    __asm__ __volatile__ ("mov %0, %0, ror %1": "=r" (val):"0" (val), "r" (shift));
      break;
    }

  /* Adjust the return value for Rn == program counter.
     PCOffset = 8 if Rn == PC, otherwise it is 0. */
  val += PCOffset;

  return val;
}

CORE_ADDR 
read_memory_opcode (CORE_ADDR mem_addr)
{
  CORE_ADDR mem_pc = mem_addr;
  if (IN_SOLIB_CALL_TRAMPOLINE (mem_addr, 0))
    {
      mem_pc = SKIP_TRAMPOLINE_CODE (mem_addr);
      if (0 == mem_pc)
	mem_pc = mem_addr;
    }

  if (read_memory (mem_pc, &mem_pc, 4) < 0)
    mem_pc = read_pc () + 4;	/* not valid */
  else
    mem_pc = ADDR_BITS_REMOVE (mem_pc);

  return (mem_pc);
}

/* Read the state of the carry flag from the PSR. */
static unsigned long
status_getcarry (void)
{
  return (read_register (PS_REGNUM) & FLAG_C ? 1 : 0);
}

/* FIXME: The following routine does not handle an SWI instruction
   correctly. */
/* Determine the branch address for an ARM instruction. */
CORE_ADDR 
branch_destination (unsigned long opcode)
{
  CORE_ADDR alt = read_pc () + 4;

  switch (opcode & 0x0e100000)
    {
      /*
         test for ARM data processing instructions
         mask = 0000 1110 0001 0000 0000 0000 0000 0000

         Look for 3 bits indicating data processing instruction and
         whether the S bit is set (condition codes are updated if 
         the S bit is set.
       */
    case 0x00000000:		/* data processing immediate/register shift */
    case 0x00100000:		/* data processing immediate */
    case 0x02000000:		/* data processing immediate/register shift */
    case 0x02100000:		/* data processing immediate */
      {
	/*
	   test for opcode and Rd = program counter in the opcode
	   mask = 0000 0001 1110 0000 1111 0000 0000 0000
	 */
	switch (opcode & 0x01e0f000)
	  {
	  case 0x0000f000:	/* and */
	    alt = opcode_getrn (opcode) & opcode_getaluop2 (opcode);
	    break;

	  case 0x0020f000:	/* eor */
	    alt = opcode_getrn (opcode) ^ opcode_getaluop2 (opcode);
	    break;

	  case 0x0040f000:	/* sub */
	    alt = opcode_getrn (opcode) - opcode_getaluop2 (opcode);
	    break;

	  case 0x0060f000:	/* rsb */
	    alt = opcode_getaluop2 (opcode) - opcode_getrn (opcode);
	    break;

	  case 0x0080f000:	/* add */
	    alt = opcode_getrn (opcode) + opcode_getaluop2 (opcode);
	    break;

	  case 0x00a0f000:	/* adc */
	    alt = opcode_getrn (opcode) + opcode_getaluop2 (opcode);
	    alt += status_getcarry ();
	    break;

	  case 0x00c0f000:	/* sbc */
	    alt = opcode_getrn (opcode) - opcode_getaluop2 (opcode);
	    alt += status_getcarry ();
	    break;

	  case 0x00e0f000:	/* rsc */
	    alt = opcode_getaluop2 (opcode) - opcode_getrn (opcode);
	    alt += status_getcarry ();
	    break;

	  case 0x0100f000:	/* tst */
	  case 0x0120f000:	/* teq */
	  case 0x0140f000:	/* cmp */
	  case 0x0160f000:	/* cmn */
	    /* these instructions never affect the PC. */
	    break;

	  case 0x0180f000:	/* orr */
	    alt = opcode_getrn (opcode) | opcode_getaluop2 (opcode);
	    break;

	  case 0x01a0f000:	/* mov */
	    alt = opcode_getaluop2 (opcode);
	    break;

	  case 0x01c0f000:	/* bic */
	    alt = opcode_getrn (opcode) & ~opcode_getaluop2 (opcode);
	    break;

	  case 0x01e0f000:	/* mvn */
	    alt = ~opcode_getaluop2 (opcode);
	    break;
	  }
      }
      break;

    case 0x06100000:		/* ldr register offset, ldr scaled register offset */
      {
	if (opcode_is_dest_pc (opcode))
	  {
	    /* get the address from Rn */
	    alt = opcode_getrn (opcode);

	    /* test the P bit */
	    if (opcode & (1 << 24))
	      {
		/* test the W bit to determine + or - */
		if (opcode & (1 << 23))
		  alt += opcode_getldrop2 (opcode);
		else
		  alt -= opcode_getldrop2 (opcode);
	      }

	    alt += 8;
	    alt = read_memory_opcode (alt);
	  }
      }
      break;

    case 0x04100000:		/* ldr imm */
      {
	if (opcode_is_dest_pc (opcode))
	  {
	    alt = opcode_getrn (opcode);
	    if (opcode & 0x00000001 << 24)
	      {
		if (opcode & 0x00000001 << 23)
		  alt += opcode & 0x00000fff;
		else
		  alt -= opcode & 0x00000fff;
	      }

	    alt = read_memory_opcode (alt);
	  }
      }
      break;

    case 0x08100000:		/* ldm */
      {
	if (opcode & (1 << 15))
	  {
	    unsigned long base;
	    int nr_regs;

	    if (opcode & (1 << 23))
	      {
		nr_regs = opcode & 65535;
		nr_regs = (nr_regs & 0x5555) + ((nr_regs & 0xaaaa) >> 1);
		nr_regs = (nr_regs & 0x3333) + ((nr_regs & 0xcccc) >> 2);
		nr_regs = (nr_regs & 0x0707) + ((nr_regs & 0x7070) >> 4);
		nr_regs = (nr_regs & 0x000f) + ((nr_regs & 0x0f00) >> 8);
		nr_regs <<= 2;

		if (!(opcode & (1 << 24)))
		  nr_regs -= 4;
	      }
	    else
	      {
		if (opcode & (1 << 24))
		  nr_regs = -4;
		else
		  nr_regs = 0;
	      }

	    base = opcode_getrn (opcode);

	    alt = read_memory_opcode (base + nr_regs);

#if 0
	    if (read_memory (base + nr_regs, &alt, 4) < 0)
	      alt = read_pc () + 4;	/* not valid */
	    else
	      alt = ADDR_BITS_REMOVE (alt);
#endif
	  }
      }
      break;

    case 0x0a000000:		/* b  */
    case 0x0a100000:		/* bl */
      {
	/* It's a branch or branch and link opcode.

	   Instead of trying to determine if the branch will be taken or not,
	   simply put a breakpoint at either location.  This is simpler, more
	   reliable, and probably not a whole lot slower than the alternative
	   approach of emulating the branch to determine its destination.
	 */
	signed long displ;

	displ = (opcode & 0x00ffffff) << 8;
	displ = (displ >> 6) + 8;
	if (displ != 0 && displ != 4)
	  alt = read_pc () + displ;
      }
      break;
    }				/* end switch (opcode & 0x0e100000) */

  return alt;
}

/* single_step() is called just before we want to resume the inferior,
   if we want to single-step it but there is no hardware or kernel
   single-step support.  We find all the possible targets of the
   coming instruction and breakpoint them.

   single_step is also called just after the inferior stops.  If we had
   set up a simulated single-step, we undo our damage.  */

void 
single_step (signal)
     enum target_signal signal;
{
  int i;
  unsigned long opcode;
  static char breakpoint_insn[] = ARM_BREAKPOINT;

  if (!one_stepped)
    {
      /* get the opcode at the current PC */
      opcode = read_memory_integer (read_pc (), 4);

      /* get the PC values for possible breakpoints */
      step_break_locations[0] = read_pc () + 4;
      step_break_locations[1] = branch_destination (opcode);

      /* Don't put two breakpoints on the same address. */
      if (step_break_locations[1] == step_break_locations[0])
	step_break_locations[1] = -1;

      /* Place breakpoint on solib trampoline target, to allow stepping into
         functions in shared libraries. */
      for (i = 0; i < 2; i++)
	if (-1 != step_break_locations[i])
	  if (IN_SOLIB_CALL_TRAMPOLINE (step_break_locations[i], 0))
	    {
	      CORE_ADDR pc = SKIP_TRAMPOLINE_CODE (step_break_locations[i]);
	      if (0 != pc)
		step_break_locations[i] = pc;
	    }

      /* insert the step breakpoints */
      for (i = 0; i < 2; i++)
	{
	  /* ignore invalid breakpoint. */
	  if (-1 == step_break_locations[i])
	    continue;

	  read_memory (step_break_locations[i], &step_shadow_insns[i], 4);
	  write_memory (step_break_locations[i], breakpoint_insn, 4);
	}

      one_stepped = 1;
    }
  else
    {
      /* remove step breakpoints. */
      for (i = 0; i < 2; i++)
	if (-1 != step_break_locations[i])
	  write_memory (step_break_locations[i], &step_shadow_insns[i], 4);

      one_stepped = 0;
    }
}
#endif
